local ADDON_NAME, dssk = ...

-- ---------------
-- ListEntry class
-- ---------------
-- Wraps a database entry to provide to more friendly output.

local ListEntry = {}

ListEntry.mt = {
  __index = function(list_entry, key)
    if key == "map_name" then
      local map_id = list_entry.db_entry.key_map_id
      return map_id and dssk.MAPS_BY_ID[map_id] and
        dssk.MAPS_BY_ID[map_id].name or nil
    elseif key == "class_name" then
      list_entry.class_name = list_entry.sex == 3 and
        LOCALIZED_CLASS_NAMES_FEMALE[list_entry.class] or
        LOCALIZED_CLASS_NAMES_MALE[list_entry.class]
      return list_entry.class_name
    elseif key == "reward_level" then
      local reward = list_entry.dungeon_reward
      if reward and reward[1] then
        return reward[1]
      else
        return nil
      end
    end

    if list_entry.db_entry[key] ~= nil then
      return list_entry.db_entry[key]
    else
      return list_entry.unit_info[key]
    end
  end
}

function ListEntry.from_db_entry(db_entry, unit_info)
  local list_entry = {}
  list_entry.db_entry = db_entry
  list_entry.unit_info = unit_info or {}
  setmetatable(list_entry, ListEntry.mt)
  return list_entry
end

-- --------------
-- SortFunc class
-- --------------
-- Hides some of the complexity of the different ways we can sort
-- the list of characters.

local SortFunc = {}
SortFunc.prototype = {
  inverted = false,
  default_inverted = false,
  nil_high = false,
}

function SortFunc.prototype:compare(left, right)
  local ret = self:f(left[self.attr], right[self.attr])
  if self.inverted then
    ret = ret * -1
  end
  return ret
end

function SortFunc.prototype:f(left, right)
  if left == nil then
    if right == nil then
      return 0
    else
      return self.nil_high and -1 or 1
    end
  end
  if right == nil then
    return self.nil_high and 1 or -1
  end
  if left < right then
    return 1
  elseif right < left then
    return -1
  end
  return 0
end

SortFunc.mt = {
  __index = SortFunc.prototype
}

function  SortFunc.new(o)
  if o.inverted and o.default_inverted == nil then
    o.default_inverted = true
  end
  o.name = o.name or o.attr
  setmetatable(o, SortFunc.mt)
  return o
end

local function reward_compare(sort_func, left, right)
  local ret = 0
  if left.reward_available then
    ret = - 1
  end
  if right.reward_available then
    ret = ret + 1
  end
  if ret == 0 then
    if (left.reward_level or 0) < (right.reward_level or 0) then
      ret = 1
    elseif (right.reward_level or 0) < (left.reward_level or 0) then
      ret = -1
    end
  end
  if sort_func.inverted then
    ret = ret * -1
  end
  return ret
end

-- --------------------------------------
-- List of characters prepared for output
-- --------------------------------------

local CharacterList = {}

CharacterList.prototype = {
  clear = function(self)
    self.list = {}
    self.by_guid = {}
  end,

  -- add a character to the list
  add = function(self, db_entry, short_guid)
    local unit_info = dssk:guid_info(short_guid)
    local list_entry = ListEntry.from_db_entry(db_entry, unit_info)
    table.insert(self.list, list_entry)
    -- by short-form guid
    self.by_guid[short_guid] = list_entry
    self.dirty = true
  end,

  -- get character by short-form GUID.
  get = function(self, short_guid)
    return self.by_guid[short_guid]
  end,

  -- get sorted list of characters.
  get_list = function(self)
    if self.dirty then
      table.sort(self.list, self.static_compare_func)
    end
    return self.list
  end,

  compare_func = function(self, left, right)
    for _, f in ipairs(self.sort_stack) do
      local result = f:compare(left, right)
      if result > 0 then return true end
      if result < 0 then return false end
    end
    return false
  end,

  sort_by = function(self, col)
    local sort_stack = self.sort_stack
    if col == sort_stack[1] then
      -- if current sort, change direction
      col.inverted = not col.inverted
    else
      -- move to the top of the list
      col.inverted = col.default_inverted
      for i = #sort_stack, 2, -1 do
        if sort_stack[i] == col then
          sort_stack[i] = sort_stack[i-1]
          -- inefficient, but concise
          sort_stack[i-1] = col
        end
      end
    end
    self.dirty = true
  end,
}

CharacterList.mt = {
  __index = CharacterList.prototype,
}

function CharacterList.new()
  local manager = {
    -- Sort order, set to default
    sort_stack = {
      SortFunc.new{ attr="faction", inverted=true },
      SortFunc.new{ attr="key_level", inverted=true },
      SortFunc.new{ name="reward", inverted=true, compare=reward_compare },
      SortFunc.new{ attr="name" },
      SortFunc.new{ attr="map_name", nil_high=true },
      SortFunc.new{ attr="realm" },
      SortFunc.new{ attr="class" },
    },
    -- sortable columns to pass to set_sort
    col = {},
  }
  function manager.static_compare_func(left, right)
    return manager:compare_func(left, right)
  end
  for _, v in pairs(manager.sort_stack) do
    manager.col[v.name] = v
  end
  setmetatable(manager, CharacterList.mt)
  manager:clear()
  return manager 
end

dssk.my_char_list = CharacterList.new()
dssk.guild_char_list = CharacterList.new()
