local ADDON_NAME, dssk = ...

dssk.chat_cooldown = {}
dssk.last_bang_keys = {}

function dssk:get_affixes(level)
  if not self.affixes then
    self.affixes = C_MythicPlus.GetCurrentAffixes()
    -- affixes still not found, request them
    if not self.affixes then
      C_MythicPlus.RequestCurrentAffixes()
      return 0, 0, 0, 0
    end
  end
  return
    level > 1 and self.affixes[1].id or 0,
    level > 3 and self.affixes[2].id or 0,
    level > 6 and self.affixes[3].id or 0,
    level > 9 and self.affixes[4].id or 0
end

-- return current week number and time within that week
local function get_curr_week_time()
  local lt = C_DateAndTime.GetServerTimeLocal()
  local t = GetServerTime()
  local diff = lt - t
  if diff > 0 and diff < dssk.MAX_EU_UTC_OFFSET then
    t = t - dssk.BASE_RESET_TIME_EU
  else
    t = t - dssk.BASE_RESET_TIME_OTHER
  end
  return math.floor(t / dssk.WEEK_SECONDS), t % dssk.WEEK_SECONDS
end

-- return a keystone item link for the provided map and level
function dssk:make_keystone_link(mapid, level)
  local map_name = C_ChallengeMode.GetMapUIInfo(mapid)
  local a1, a2, a3, a4 = self:get_affixes(level)
  return string.format(
    "|c%s|Hkeystone:%s:%s:%s:%s:%s:%s:%s|h[Keystone: %s (%s)]|h|r",
    self.KEYSTONE_LINK_COLOUR,
    self.KEYSTONE_ITEM_ID,
    mapid,
    level,
    a1, a2, a3, a4,
    map_name,
    level)
end

local cached_realm_name = nil
function dssk:get_realm_name()
  if not cached_realm_name then
    cached_realm_name = gsub(GetRealmName(), " ", "")
  end
  return cached_realm_name
end

-- stricter keystone link extraction for use with chat messages
-- returns map_id and key_level if the entire message is a valid keystone link
-- otherwise returns nil
function dssk:is_keystone_link(msg_str)
  if not msg_str then return nil end
  local map_id, key_level = strmatch(msg_str, self.KEYSTONE_MATCH_PATTERN)
  return tonumber(map_id), tonumber(key_level)
end

-- return map_id and level of provided keystone item link
function dssk:link_to_map_level(item_link)
  if not item_link then
    return
  end
  -- first try keystone style item link
  local map_id, level = self:is_keystone_link(item_link)
  if not map_id then
    -- failed, so try item style link
    local id_count, data = strmatch(item_link, self.KEYSTONE_ITEM_MATCH_PATTERN)
    -- search through "bonus id" section for map and level data
    if id_count then
      id_count = tonumber(id_count)
      -- note: this pattern will be brittle if empty bonus IDs are a thing.
      for id, value in gmatch(data, "(%d+):(%d+)") do
        if id_count == 0 then
          break
        end
        id_count =  id_count - 1
        if id == "17" then
          map_id = tonumber(value)
        elseif id == "18" then
          level = tonumber(value)
        end
      end
    end
  end
  return map_id, level
end

-- search bag for a keystone, return bag_index, slot and itemlink
function dssk:find_key_in_bag()
  for bag = 0, NUM_BAG_SLOTS do
    for slot = 1, GetContainerNumSlots(bag) do
      if GetContainerItemID(bag, slot) == self.KEYSTONE_ITEM_ID then
        return bag, slot, GetContainerItemLink(bag, slot)
      end
    end
  end
  return nil
end

local DEFAULT_SETTINGS = {
  auto_slot = true,
  announce_new_key = true,
  chat = {
    party = { reply = true, alts = false },
    guild = { reply = true, alts = false },
    raid = { reply = false, alts = false },
  },
  key_sync = true,
}

-- set all settings to defaults
function dssk:default_settings()
  for k,v in pairs(DEFAULT_SETTINGS) do
    self.settings[k] = v
  end
end

-- load saved variables into addon namespace
function dssk:load_vars()
  DidSomebodySayKeys_DB = DidSomebodySayKeys_DB or {}
  DidSomebodySayKeys_Settings = DidSomebodySayKeys_Settings or {}
  self.db = DidSomebodySayKeys_DB
  self.settings = DidSomebodySayKeys_Settings

  for k, v in pairs(DEFAULT_SETTINGS) do
    if self.settings[k] == nil then
      self.settings[k] = v
    end
  end

  self:load_guid_cache()

  if self.db.my_characters then
    for guid, db_entry in pairs(self.db.my_characters) do
      self.my_char_list:add(db_entry, guid)
    end
  end
end

function dssk:load_guild(guild_name)
  local db = self:check_guild_db(guild_name)
  if db then
    for guid, db_entry in pairs(db) do
      self.guild_char_list:add(db_entry, guid)
    end
  end
end

function dssk:reset_db()
  for k, v in pairs(self.db) do
    self.db[k] = nil
  end

  self.my_char_list:clear()
  self.guild_char_list:clear()

  self.db.version = GetAddOnMetadata(ADDON_NAME, "Version")
  self.db.week = get_curr_week_time() 
  self.db.my_characters = {}
  self.db.guilds = {}
end

function dssk:rollover_db()
  local chars = self.db.my_characters
  for k, db_entry in pairs(chars) do
    if db_entry.dungeon_reward and db_entry.dungeon_reward[1]
        and db_entry.dungeon_reward[1] > 0 then
      -- delete old keystone
      db_entry.key_level = nil
      db_entry.key_map_id = nil
      -- mark reward available
      db_entry.reward_available = true
    else
      -- delete characters without rewards
      chars[k] = nil
    end
  end

  self.my_char_list:clear()
  self.guild_char_list:clear()

  -- re-add characters to display list
  for guid, db_entry in pairs(chars) do
    self.my_char_list:add(db_entry, guid)
  end

  self.db.version = GetAddOnMetadata(ADDON_NAME, "Version")
  self.db.week = get_curr_week_time() 
  self.db.guilds = {}
end

-- create empty database if it doesn't exist or is from a previous week
function dssk:check_db()
  local week = get_curr_week_time()

  if not self.db.version or not self.db.week then
    self:reset_db()
    C_MythicPlus.RequestCurrentAffixes()
    return
  end

  -- parse version
  -- consider patch differences to be database compatible
  -- minor or major difference should reset the database
  local major, minor, patch = strsplit(".", self.db.version)
  if major ~= self.VERSION.major or minor ~= self.VERSION.minor then
    self:reset_db()
    C_MythicPlus.RequestCurrentAffixes()
    return
  end

  if self.db.week ~= week then
    if self.db.week == week - 1 then
      -- last week's data, do a rollover
      self:rollover_db()
    else
      -- old data, throw it away
      self:reset_db()
    end
    -- request new affixes on reset
    C_MythicPlus.RequestCurrentAffixes()
  end
end

function dssk:check_guild_db(guild)
  self:check_db()
  if not self.db.guilds[guild] then
    self.db.guilds[guild] = {}
    self.guild_char_list:clear()
  end
  return self.db.guilds[guild]
end

-- copy current player details (but not key) to saved data
function dssk:update_player_db_entry()
  if not self.player or self.player_level < self.REQUIRED_LEVEL then
    return nil
  end

  self:check_db()

  local chars = self.db.my_characters
  -- index by short guid
  local guid = self.player_short_guid
  local entry = chars[guid]
  local entry_is_new = not entry

  if entry_is_new then
    entry = {}
    chars[guid] = entry
  end
  for k, v in pairs(self.player) do
    entry[k] = v
  end
  if entry_is_new then
    self.my_char_list:add(entry, guid)
  end

  return entry
end

function dssk:player_login()
  self.player = {
    faction = UnitFactionGroup("player"),
  }
  -- level stored outside player, because we don't need to save it
  self.player_name = UnitName("player")
  self.player_level = UnitLevel("player")
  self.player_guid = UnitGUID("player")
  self.player_short_guid = self:squish_player_guid(self.player_guid)
  self:capture_guid(self.player_guid)

  local db_entry = self:update_player_db_entry()

  C_MythicPlus.RequestMapInfo()
  if not self.affixes then
    C_MythicPlus.RequestCurrentAffixes()
  end
  C_MythicPlus.RequestRewards()

  -- ping this incase they change the login/guild update order
  self:player_guild_update()

  local key_map = C_MythicPlus.GetOwnedKeystoneChallengeMapID()
  if key_map then
    local key_level = C_MythicPlus.GetOwnedKeystoneLevel()
    self:set_my_key(key_map, key_level)
  end
end

function dssk:set_my_key(map_id, key_level, announce)
  local db_entry = self:update_player_db_entry()
  if db_entry then
    local new_key = db_entry.key_level ~= key_level or
      db_entry.key_map_id ~= map_id
    if not new_key then
      announce = false
    else
      announce = announce and self.settings.announce_new_key
      db_entry.key_level = key_level
      db_entry.key_map_id = map_id
    end

    -- add a copy to guild keys too
    if self.player.guild then
      self:add_guild_key_from_guid(
        self.player_guid,
        map_id,
        key_level
      )
    end
    if new_key then
      self:send_my_key()
    end
    if announce and GetNumGroupMembers() > 1 then
      self:report_a_key('party', false, db_entry, true)
    end
  end
end

function dssk:update_key(announce)
  local _, _, key_link = self:find_key_in_bag()

  if key_link then
    local map_id, level = self:link_to_map_level(key_link)
    self:set_my_key(map_id, level, announce)
  end
end

function dssk:update_rewards()
  local db_entry = self:update_player_db_entry()
  if db_entry then
    db_entry.reward_available = C_WeeklyRewards.HasAvailableRewards()
    -- preserve any existing reward level while rewards available
    if not db_entry.reward_available then
      local activities = C_WeeklyRewards.GetActivities(1)
      db_entry.dungeon_reward = {}
      for _, v in pairs(activities) do
        db_entry.dungeon_reward[v.index] = v.level or 0
      end
      activities = C_WeeklyRewards.GetActivities(3)
      db_entry.raid_reward = {}
      for _, v in pairs(activities) do
        db_entry.raid_reward[v.index] = v.level or 0
      end
    end
  end
end

function dssk:add_guild_key_from_guid(guid, map_id, key_level)
  self:capture_guid(guid)
  local short_guid = self:squish_player_guid(guid)
  self:add_guild_key_from_short_guid(short_guid, map_id, key_level)
end

function dssk:add_guild_key_from_short_guid(short_guid, map_id, key_level)
  local db = self:check_guild_db(self.player.guild)
  local entry = db[short_guid]
  local entry_is_new = not entry
  if entry_is_new then
    entry = {}
    db[short_guid] = entry
  end
  entry.key_map_id = map_id
  entry.key_level = key_level
  if entry_is_new then
    self.guild_char_list:add(entry, short_guid)
  end
end

function dssk:report_keys(channel, anonymous)
  if not self.settings.chat[channel].reply then
    return
  end
  local now = GetServerTime()
  if now < (self.chat_cooldown[channel] or 0) then
    return
  end
  self.chat_cooldown[channel] = now + self.CHAT_COOLDOWN_SECONDS

  self:check_db()

  local chars = self.my_char_list:get_list()
  -- current char first
  local cur_char = self.my_char_list:get(self.player_short_guid)
  self:report_a_key(channel, anonymous, cur_char, true)
  if self.settings.chat[channel].alts and self.player.guild then
    for _, v in pairs(chars) do
      if v ~= cur_char and v.guild == self.player.guild then
        self:report_a_key(channel, anonymous, v, false)
      end
    end
  end
end

function dssk:report_a_key(channel, anonymous, entry, current)
  if entry and entry.key_level then
    local link = self:make_keystone_link(entry.key_map_id, entry.key_level)
    if current then
      SendChatMessage(link, channel)
    else
      local name
      if anonymous then
        name = "On " .. entry.class_name .. " alt"
      else
        name = string.format("%s (%s)", entry.name, entry.class_name)
      end
      SendChatMessage(string.format("%s -%s-", name, link), channel)
    end
  end
end

function dssk:scrape_chat_key(msg, channel, sender_name, sender_guid)
  -- test if key link
  local map_id, key_level = self:is_keystone_link(msg)
  if map_id and key_level then
    -- we only want to log keys for our guild members
    if channel ~= "guild" then
      if not IsGuildMember(sender_guid) then
        return
      end
    end
    -- add key to list
    self:add_guild_key_from_guid(sender_guid, map_id, key_level)
  end
end

function dssk:handle_chat(msg, channel, sender_name, sender_guid, anonymous)
  if IsGuildMember(sender_guid) then
    self:capture_guid(sender_guid)
  end
  if msg == "!keys" then
    self.last_bang_keys[channel] = GetServerTime()
    self:report_keys(channel, anonymous)
  else
    local last_request = self.last_bang_keys[channel]
    if last_request and self.player and self.player.guild then
      local since = GetServerTime() - last_request
      if since < dssk.BANG_KEYS_LISTEN_SECONDS then
        self:scrape_chat_key(msg, channel, sender_name, sender_guid)
      end
    end
  end
end

-- current player's guild (possibly) changed
function dssk:player_guild_update()
  if not self.player then
    return
  end
  if not IsInGuild() then
    self.player.guild = nil
    self:update_player_db_entry()
    self.guild_char_list:clear()
  else
    local gname, _, _, realm = GetGuildInfo('player')
    if gname then
      if not realm then
        realm = self:get_realm_name()
      end
      local sqgn = string.format("%s-%s", gname, realm)
      if self.player.guild ~= sqgn then
        self.player.guild = sqgn
        self:load_guild(sqgn)
        -- ping our key to add it to guild keys on login
        self:update_key()
        -- setup background key syncing
        self:init_addon_chat()
      else
        self:update_player_db_entry()
      end
    end
  end
end

function dssk:player_level_update(level)
  self.player_level = level
  self:update_player_db_entry()
end


function dssk:key_looted(player_name, item_link)
  if self.player then
    if player_name == self.player_name then
      self:set_my_key(self:link_to_map_level(item_link))
      -- this could be our cache, request rewards update
      C_MythicPlus.RequestRewards()
    else
      if self.player.guild then
        local guid = UnitGUID(player_name)
        if not IsGuildMember(guid) then
          return
        end
        -- add key to list
        local map_id, key_level = self:link_to_map_level(item_link)
        if map_id and key_level then
          self:add_guild_key_from_guid(guid, map_id, key_level)
        end
      end
    end
  end
end

function dssk:toggle_window()
  if InCombatLockdown() then
    -- Can't use the UIPanel API in combat.
    -- Need a better work around than this.
    self.frame:SetShown(not self.frame:IsShown())
  else
    if self.frame:IsShown() then
      HideUIPanel(self.frame)
    else
      ShowUIPanel(self.frame)
    end
  end
end
